{ 4IN1ROWW.PAS - Four in a row (Windows game)

  title    : 4IN1ROWW
  version  : 1.2
  date     : Feb 10, 2000
  language : Borland Pascal v7.0 with Object Windows
  usage    : Windows v3.1 application
  author   : J R Ferguson
  e-mail   : j.r.ferguson@iname.com
  download : http://hello.to/ferguson

  This program and its source may be used and copied freely whithout
  charge, but only for non-commercial purposes. In no way the author 
  can be held responsible for any damage or loss of data that may be
  caused by using this software.
}

{$B-}  { short circuit Boolean expression evaluation }
{$V-}  { relaxed var-string checking }
{$X+}  { extended syntax }

{$UNDEF  DEBUG}  { add debug code }

program VIERRIJW;

uses WinTypes, WinProcs, Objects, OWindows, ODialogs, Strings;

{$R 4IN1ROWW.RES}
{$I 4IN1ROWW.INC}

type
  T_Player      = (Play_Red,          { who plays (computer against red or two players) }
                   Play_Two);
  T_WhoStarts   = (Start_Red,         { who starts the game }
                   Start_Blue,
                   Start_Random);
  T_Figure      = (Fig_Table,         { figures with different drawing pencils and brushes}
                   Fig_Board,
                   Fig_Field,
                   Fig_Red,
                   Fig_Blue,
                   Fig_Mark);
  T_ScanResult  = (Scn_Normal,
                   Scn_Full,
                   Scn_PrevWins,
                   Scn_NextWins);

const
{ Constant values }
  C_WinLen      = 4;                  { length of a winning row }
  C_MinColumns  = 5;                  { minimum number of vertical columns }
  C_MaxColumns  = 9;                  { maximum number of vertical columns }
  C_MaxRows     = 6;                  { maximum number of horizontal rows }
  C_RedPlaysStr = ' - red plays';
  C_BluePlaysStr= ' - blue plays';
  C_GameOverStr = ' - game over';
  C_Weight      : array[boolean,1..C_WinLen-1] of integer =
                  ( (1, 50,6666),     { next move (self) }
                    (1,100,9999) );   { current move (opponent) }

{ Default values }
  C_DflColumns  = 7;                  { number of vertical columns (C_MinColumns..C_MaxColumns) }
  C_DflPlayer   = Play_Red;           { who plays (T_Player) }
  C_DflFirstMove= Start_Red;          { who plays first (T_WhoStarts) }

type
  P_DrawSet     = ^T_DrawSet;         { pencil and brush to draw a figure }
  P_DrawTools   = ^T_DrawTools;       { collection of all drawsets needed }
  P_GameObject  = ^T_GameObject;      { base object type for picture elements }
  P_Mark        = ^T_Mark;            { mark on a disc  }
  P_Disc        = ^T_Disc;            { blue or red disc to play }
  P_Field       = ^T_Field;           { potential place for a disc }
  P_Board       = ^T_Board;           { game board with fields }
  P_OptionsBuf  = ^T_OptionsBuf;      { options dialog data buffer  }
  P_OptionsDlg  = ^T_OptionsDlg;      { options dialog }
  P_GameOverBuf = ^T_GameOverBuf;     { game over dialog data buffer }
  P_GameOverDlg = ^T_GameOverDlg;     { game over dialog }
  P_Game        = ^T_Game;            { application window }
  P_Application = ^T_Application;     { Windows application }

  T_DrawSet     = object(TObject)
    Pen         : HPen;
    Brush       : HBrush;
    constructor Init(V_LineWidth: integer; V_PenColor, V_BrushColor: longint);
    destructor  Done; virtual;
  end;

  T_DrawTools   = object(TObject)
    DrawSet     : array[T_Figure] of P_DrawSet;
    constructor Init;
    destructor  Done; virtual;
    procedure   Select(V_Context: HDC; V_Figure: T_Figure);
  end;

  T_GameObject  = object(TObject)
    X, Y        : integer;             { pixel position of center }
    Ax, Ay      : integer;             { size (half side) horizontal/vertical }
    Figure      : T_Figure;            { figure type }
    constructor Init(V_Figure: T_Figure);
    procedure   Draw(V_Context: HDC; V_X, V_Y, V_Ax, V_Ay: integer); virtual;
  end;

  T_Mark        = object(T_GameObject)
  end;

  T_Disc        = object(T_GameObject)
    Mark        : P_Mark;
    constructor Init(V_Figure: T_Figure);
    destructor  Done; virtual;
    procedure   DoMark;
    procedure   Draw(V_Context: HDC; V_X, V_Y, V_Ax, V_Ay: integer); virtual;
  end;

  T_Field       = object(T_GameObject)
    Row, Column : integer;          { row and column on the board }
    Disc        : P_Disc;           { optional disc or nil }
    constructor Init(V_Row: integer; V_Col: integer);
    destructor  Done; virtual;
    procedure   Draw(V_Context: HDC; V_X, V_Y, V_Ax, V_Ay: integer); virtual;
    function    Contains(V_X, V_Y: integer): boolean;
    procedure   SetDisc(V_Red: boolean);
    procedure   DoMark;
  end;

  T_Board       = object(T_GameObject)
    MaxCol      : integer;            { number of fields horizontal }
    MaxRow      : integer;            { number of fields vertical   }
    Field       : array[1..C_MaxColumns,1..C_MaxColumns] of P_Field;
    constructor Init(V_MaxCol: integer);
    destructor  Done; virtual;
    procedure   Draw(V_Context: HDC; V_X, V_Y, V_Ax, V_Ay: integer); virtual;
    function    FullBoard: boolean;
    function    FreeField(V_X, V_Y: Integer; var V_Col: integer): boolean;
    function    FreeRow(V_Col: integer; var V_Row: integer): boolean;
    procedure   DoMove(V_Col: integer; V_Red: boolean);
    function    Scan(V_Player: T_Player; V_RedPlays: boolean): T_ScanResult;
  end;

  T_OptionsBuf  = record
    ColumnCount : array[C_MinColumns..C_MaxColumns] of word;
    Player      : array[Play_Red..Play_Two] of word;
    WhoStarts   : array[Start_Red..Start_Random] of word;
  end;

  T_OptionsDlg  = object(TDialog)
    constructor Init(V_Game: PWindowsObject; V_Title: PChar; V_OptionsBuf: P_OptionsBuf);
  end;

  T_GameOverBuf = record
    Winner      : array[0..12] of char;
  end;

  T_GameOverDlg = object(TDialog)
    constructor Init(V_Game: PWindowsObject; V_Title: PChar; V_GameOverBuf: P_GameOverBuf);
  end;

  T_Game        = object(TWindow)
    ColumnCount : integer;          { number of columns on the board }
    Player      : T_Player;         { computer against red or two players }
    WhoStarts   : T_WhoStarts;      { who starts the game (red, blue or at random) }
    Board       : P_Board;          { Game Board }
    RedPlays    : boolean;          { red plays next (yes/no) }
    BaseTitle   : PChar;            { fixed part of game window title }
    GameOver    : boolean;          { game over (yes/no) }
{$IFDEF DEBUG}
    DebugMode   : boolean;          { Debug mode }
{$ENDIF}
    constructor Init(V_Parent: PWindowsObject; V_Title: PChar);
    destructor  Done; virtual;
    procedure   SetupGame;
    procedure   ClearGame;
    procedure   NewGame;
    procedure   SetUpWindow; virtual;
    function    GetClassName: PChar; virtual;
    procedure   GetWindowClass(var V_Class: TWndClass); virtual;
    procedure   Paint(V_Context: HDC; var V_PaintInfo: TPaintStruct); virtual;
    procedure   GameOverMessage(V_Winner: boolean);
    procedure   Instruct;
    procedure   FirstMove;
    procedure   SwapPlayer;
    procedure   DoMove(V_Col: integer);
    procedure   WMLButtonDown (var V_Msg: TMessage); virtual wm_First + wm_LButtonDown;
    procedure   CMGameOptions (var V_Msg: TMessage); virtual cm_First + cm_GameOptions;
    procedure   CMGameNew     (var V_Msg: TMessage); virtual cm_First + cm_GameNew    ;
    procedure   CMHelpAbout   (var V_Msg: TMessage); virtual cm_First + cm_HelpAbout  ;
    procedure   CMHelpRules   (var V_Msg: TMessage); virtual cm_First + cm_HelpRules  ;
{$IFDEF DEBUG}
    procedure   CMDebug       (var V_Msg: TMessage); virtual cm_First + cm_Debug      ;
{$ENDIF}
  end;

  T_Application = object(TApplication)
    DrawTools   : P_DrawTools;
    constructor Init(V_Title: PChar);
    destructor  Done; virtual;
    procedure   InitMainWindow; virtual;
    procedure   InitInstance; virtual;
  end;


{ --- General --- }

function Imin(i,j: integer): integer;
begin if i<j then Imin:= i else Imin:= j; end;


{ --- T_DrawSet --- }

constructor T_DrawSet.Init(V_LineWidth: integer; V_PenColor, V_BrushColor: longint);
begin
  inherited Init;
  Pen  := CreatePen(ps_Solid, V_LineWidth, V_PenColor);
  Brush:= CreateSolidBrush(V_BrushColor);
end;

destructor  T_DrawSet.Done;
begin
  DeleteObject(Pen);
  DeleteObject(Brush);
  inherited Done;
end;


{ --- T_DrawTools --- }

constructor T_DrawTools.Init;
begin
  inherited Init;
  New(DrawSet[Fig_Table],Init(0,RGB(255,255,255),RGB(255,255,255)));
  New(DrawSet[Fig_Board],Init(0,RGB(  0,  0,  0),RGB(127, 64,  0))); {not used}
  New(DrawSet[Fig_Field],Init(0,RGB(  0,  0,  0),RGB(192,192,192)));
  New(DrawSet[Fig_Red  ],Init(0,RGB(255,  0,  0),RGB(255,  0,  0)));
  New(DrawSet[Fig_Blue ],Init(0,RGB(  0,  0,255),RGB(  0,  0,255)));
  New(DrawSet[Fig_Mark ],Init(2,RGB(255,255,  0),RGB(  0,  0,  0)));
end;

destructor  T_DrawTools.Done;
begin
  Dispose(DrawSet[Fig_Table],Done);
  Dispose(DrawSet[Fig_Board],Done);
  Dispose(DrawSet[Fig_Field],Done);
  Dispose(DrawSet[Fig_Red  ],Done);
  Dispose(DrawSet[Fig_Blue ],Done);
  Dispose(Drawset[Fig_Mark ],Done);
  inherited Done;
end;

procedure   T_DrawTools.Select(V_Context: HDC; V_Figure: T_Figure);
begin
  SelectObject(V_Context, DrawSet[V_Figure]^.Pen);
  SelectObject(V_Context, DrawSet[V_Figure]^.Brush);
end;


{ --- T_GameObject --- }

constructor T_GameObject.Init(V_Figure: T_Figure);
begin
  inherited Init;
  X:= 0; Y:= 0; Ax:= 0; Ay:= 0;
  Figure:= V_Figure;
end;

procedure   T_GameObject.Draw(V_Context: HDC; V_X, V_Y, V_Ax, V_Ay: integer);
begin
  X:= V_X; Y:= V_Y; Ax:= V_Ax; Ay:= V_Ay;
  P_Application(Application)^.DrawTools^.Select(V_Context,Figure);
  case Figure of
    Fig_Table : { no operation };
    Fig_Board : { no operation };
    Fig_Field : Rectangle(V_Context, X-Ax,Y-Ay, X+Ax,Y+Ay);
    Fig_Red,
    Fig_Blue  : Ellipse  (V_Context, X-Ax,Y-Ay, X+Ax,Y+Ay);
    Fig_Mark  : Arc(V_Context, X-Ax,Y-Ay, X+Ax, Y+Ay, X-Ax, Y-Ay, X-Ax, Y-Ay);
  end;
end;


{ --- T_Mark --- }



{ --- T_Disc --- }


constructor T_Disc.Init(V_Figure: T_Figure);
begin inherited Init(V_Figure); Mark:= nil; end;

destructor T_Disc.Done;
begin
  if Mark <> nil then Dispose(Mark,Done);
  inherited Done;
end;

procedure   T_Disc.DoMark;
var Area: TRect;
begin if Mark = nil then begin
  New(Mark,Init(Fig_Mark));
  with Area do begin Left:= X-Ax; Top:= Y-Ay; Right:= X+Ax; Bottom:= Y+Ay; end;
  InvalidateRect(Application^.MainWindow^.HWindow,@Area,false);
end end;

procedure   T_Disc.Draw(V_Context: HDC; V_X, V_Y, V_Ax, V_Ay: integer);
begin
  inherited Draw(V_Context, V_X, V_Y, V_Ax, V_Ay);
  if Mark <> nil then Mark^.Draw(V_Context, V_X, V_Y, V_Ax, V_Ay);
end;


{ --- T_Field --- }

constructor T_Field.Init(V_Row: integer; V_Col: integer);
begin
  inherited Init(Fig_Field);
  Row    := V_Row;
  Column := V_Col;
  Disc   := nil;
end;

destructor  T_Field.Done;
begin
  if Disc <> nil then Dispose(Disc,Done);
  inherited Done;
end;

procedure   T_Field.Draw(V_Context: HDC; V_X, V_Y, V_Ax, V_Ay: integer);
var R: integer;
begin
  inherited Draw(V_Context, V_X, V_Y, V_Ax, V_Ay);
  if Disc <> nil then begin
    R:= Imin(Ax,Ay); R:= R - R div 5;
    Disc^.Draw(V_Context, V_X, V_Y, R, R);
  end;
end;

function    T_Field.Contains(V_X, V_Y: integer): boolean;
begin Contains:= (Abs(V_X - X) <= Ax) and (Abs(V_Y - Y) <= Ay); end;

procedure   T_Field.SetDisc(V_Red: boolean);
var Area: TRect;
begin if Disc = nil then begin
  if V_Red then New(Disc,Init(Fig_Red))
           else New(Disc,Init(Fig_Blue));
  with Area do begin Left:= X-Ax; Top:= Y-Ay; Right:= X+Ax; Bottom:= Y+Ay; end;
  InvalidateRect(Application^.MainWindow^.HWindow,@Area,false);
end end;

procedure   T_Field.DoMark;
begin if Disc <> nil then Disc^.DoMark; end;


{ --- T_Board --- }

constructor T_Board.Init(V_MaxCol: integer);
var r,c: integer;
begin
  inherited Init(Fig_Board);
  MaxCol:= V_MaxCol;
  MaxRow:= Imin(C_MaxRows,MaxCol - 1);
  for r:= 1 to MaxRow do for c:= 1 to MaxCol do New(Field[r,c], Init(r,c));
end;

destructor  T_Board.Done;
var r,c: integer;
begin
  for r:= 1 to MaxRow do for c:= 1 to MaxCol do Dispose(Field[r,c],Done);
  inherited Done;
end;

procedure   T_Board.Draw(V_Context: HDC; V_X, V_Y, V_Ax, V_Ay: integer);
var X0, X1, Y0, Y1, Ax0, Ay0: integer; r,c: integer;
begin
  inherited Draw(V_Context, V_X, V_Y, V_Ax, V_Ay);
  Ax0:= Ax div MaxCol;
  Ay0:= Ay div MaxRow;
  X0 := X - Ax - Ax0;
  Y0 := Y + Ay + Ay0;
  for r:= 1 to MaxRow do for c:= 1 to MaxCol do begin
    X1:= X0 + 2 * Ax0 * c;
    Y1:= Y0 - 2 * Ay0 * r;
    Field[r,c]^.Draw(V_Context, X1, Y1, Ax0, Ay0);
  end;
end;

function    T_Board.FullBoard: boolean;
var full: boolean; c: integer;
begin
  full:= true; for c:= 1 to MaxCol do if Field[MaxRow,c]^.Disc = nil then full:= false;
  FullBoard:= full;
end;

function    T_Board.FreeField(V_X, V_Y: Integer; var V_Col: integer): boolean;
var Row: integer; found: boolean;
begin
  V_Col:= 0; found:= false;
  while (V_Col < MaxCol) and not found do begin
    Inc(V_Col); Row:= 0;
    while (Row < MaxRow) and not found do begin
      inc(Row); found:= Field[Row,V_Col]^.Contains(V_X,V_Y)
    end;
  end;
  FreeField:= found and (Field[Row,V_Col]^.Disc = nil);
end;

function    T_Board.FreeRow(V_Col: integer; var V_Row: integer): boolean;
begin
  V_Row:= MaxRow;
  while (V_Row > 1) and (Field[V_Row-1,V_Col]^.Disc = nil) do Dec(V_Row);
  FreeRow:= Field[V_Row,V_Col]^.Disc = nil;
end;

procedure   T_Board.DoMove(V_Col: integer; V_Red: boolean);
var Row: integer;
begin
  if FreeRow(V_Col,Row) then
     Field[Row,V_Col]^.SetDisc(V_Red);
end;


function    T_Board.Scan(V_Player: T_Player; V_RedPlays: boolean): T_ScanResult;
 var
   Result   : T_ScanResult;
   Col, c   : integer;
   ScanLen  : array[boolean] of integer;
   Value    : array[1..C_MaxColumns] of integer;
   WinCol   : integer;
   WinChain : array[1..C_WinLen] of record Row,Col: integer end;

  procedure DoMarkChain;
  var i: integer;
  begin for i:= 1 to C_WinLen do with WinChain[i] do Field[Row,Col]^.DoMark; end;

  function DoScan(FromRow,ToRow, FromCol,ToCol, dRow,dKol: integer): boolean;
  var Row,Col: integer; checking: boolean;

    function ScanChain: boolean;
    var i: integer; StopScan: boolean;

      procedure RememberChain;
      var i: integer;
      begin
        for i:= 1 to C_WinLen do begin
          WinChain[i].Row:= Row+(i-1)*dRow;
          WinChain[i].Col:= Col+(i-1)*dKol;
        end;
      end;

      procedure ScanMove(MoveColor: boolean);
      var i: integer; r: integer;
      begin if ScanLen[MoveColor] > 0 then begin {ScanMove}
        i:= 0;
        repeat
          if FreeRow(Col+i*dKol,r) and (Row+i*dRow = r) then begin
            inc(Value[Col+i*dKol],C_Weight[MoveColor,ScanLen[MoveColor]]);
            if (WinCol=0) and (not MoveColor) and (ScanLen[MoveColor]=C_WinLen-1) then begin
              WinCol:= Col + i*dKol; RememberChain; Result:= Scn_NextWins;
            end;
          end;
          inc(i);
        until (i = C_WinLen);
      end end;

    begin {ScanChain}
      ScanLen[false]:= 0; ScanLen[true]:= 0; StopScan:= false;
      for i:= 0 to C_WinLen - 1 do
        with Field[Row+i*dRow,Col+i*dKol]^ do if Disc <> nil then
          Inc(ScanLen[(Disc^.Figure = Fig_Red) = V_RedPlays]);
      if ScanLen[true]=C_WinLen then begin
        WinCol:= Col + i*dKol; RememberChain; Result:= Scn_PrevWins; StopScan:= true;
      end
      else if V_Player <> Play_Two then begin ScanMove(true); ScanMove(false); end;
      ScanChain:= StopScan;
    end;

  begin {DoScan}
    checking:= true; Row:= FromRow-1;
    while checking and (Row < ToRow) do begin
      Inc(Row); Col:= FromCol-1;
      while checking and (Col < ToCol) do begin
        Inc(Col); if ScanChain then checking:= false;
      end;
    end;
    DoScan:= not checking;
  end;

begin {T_Board.Scan}
  Result:= Scn_Normal; WinCol:= 0;
  for c:= 1 to MaxCol do Value[c]:= 0;
            {FromRow,ToRow     , FromCol ,ToCol            , dr,dc}
  if DoScan(1,MaxRow           , 1       ,MaxCol-C_WinLen+1,  0,+1) {horizontal}
  or DoScan(1,MaxRow-C_WinLen+1, 1       ,MaxCol           , +1, 0) {vertical}
  or DoScan(1,MaxRow-C_WinLen+1, 1       ,MaxCol-C_WinLen+1, +1,+1) {diag L-R}
  or DoScan(1,MaxRow-C_WinLen+1, C_WinLen,MaxCol           , +1,-1) {diag R-L}
  then;

  if      Result = Scn_PrevWins then DoMarkChain
  else if Result = Scn_NextWins then begin {do winning move}
    DoMove(WinCol,not V_RedPlays); DoMarkChain;
  end
  else begin
    if V_Player <> Play_Two then begin {do best move}
      Col:= 1 + Random(MaxCol);
      if Random(2)=1 then begin
        for c:= 1 to MaxCol do if Value[c] > Value[Col] then Col:= c;
      end
      else begin
        for c:= MaxCol downto 1 do if Value[c] > Value[Col] then Col:= c;
      end;
      DoMove(Col,not V_RedPlays);
    end;
    if FullBoard then Result:= Scn_Full;
  end;
  Scan:= Result;
end;



{ --- T_OptionsDlg --- }

constructor T_OptionsDlg.Init(V_Game: PWindowsObject; V_Title: PChar;
                              V_OptionsBuf: P_OptionsBuf);
var p: PControl; i: integer; s: T_Player; w: T_WhoStarts;
begin
  inherited Init(V_Game, V_Title);
  for i:= C_MinColumns to C_MaxColumns do
    p:= New(PRadioButton, InitResource(@Self, id_OptionsRows + i - C_MinColumns));
  for s:= Play_Red to Play_Two do
    p:= New(PRadioButton, InitResource(@Self, id_OptionsWhoPlays + Ord(s)));
  for w:= Start_Red to Start_Random do
    p:= New(PRadioButton, InitResource(@Self, id_OptionsWhoStarts + Ord(w)));
  TransferBuffer:= V_OptionsBuf;
end;


{ --- T_GameOverDlg --- }

constructor T_GameOverDlg.Init(V_Game: PWindowsObject; V_Title: PChar;
                               V_GameOverBuf: P_GameOverBuf);
var p: PControl;
begin
  inherited Init(V_Game, V_Title);
  p:= New(PStatic, InitResource(@Self, id_Winner, 12));
  TransferBuffer:= V_GameOverBuf;
end;


{ --- T_Game --- }

constructor T_Game.Init(V_Parent: PWindowsObject; V_Title: PChar);
begin
  inherited Init(V_Parent, V_Title);
  Attr.Menu  := LoadMenu(HInstance, MakeIntResource(MNU_MAIN));
  ColumnCount:= C_DflColumns;
  Player     := C_DflPlayer;
  WhoStarts  := C_DflFirstMove;
  BaseTitle  := StrNew(Attr.Title);
{$IFDEF DEBUG}
  DebugMode  := false;
{$ENDIF}
  SetupGame;
end;

destructor  T_Game.Done;
begin
  ClearGame;
  StrDispose(BaseTitle);
  inherited Done;
end;

procedure   T_Game.SetupGame;
begin
  case WhoStarts of
    Start_Red    : RedPlays:= true;
    Start_Blue   : RedPlays:= false;
    Start_Random : RedPlays:= Random(2) = 1;
  end;
  GameOver:= false;
  New(Board,Init(ColumnCount));
end;

procedure   T_Game.ClearGame;
begin Dispose(Board,Done); end;

procedure   T_Game.NewGame;
begin
  ClearGame; SetupGame; FirstMove;
  InvalidateRect(Application^.MainWindow^.HWindow,nil,false);
end;

procedure   T_Game.SetupWindow;
begin FirstMove; end;

function    T_Game.GetClassName: PChar;
begin GetClassName:= '4IN1ROWW'; end;

procedure   T_Game.GetWindowClass(var V_Class: TWndClass);
begin
  inherited GetWindowClass(V_Class);
  V_Class.hIcon:= LoadIcon(HInstance,MakeIntResource(ICO_MAIN));
end;

procedure   T_Game.Paint(V_Context: HDC; var V_PaintInfo: TPaintStruct);
var X,Y,Ax,Ay: integer; ClientRect: TRect;
begin
  GetClientRect(HWindow,ClientRect);
  with ClientRect do begin
    X := (Right - Left) div 2;
    Y := (Bottom - Top) div 2;
    Ax:= ColumnCount * (Imin(Right - Left, Bottom - Top) div (2 * ColumnCount + 1));
    Ay:= (Ax div Board^.MaxCol) * Board^.MaxRow;
    P_Application(Application)^.DrawTools^.Select(V_Context, Fig_Table);
    Rectangle(V_Context, Left, Top , X-Ax , Bottom);
    Rectangle(V_Context, X+Ax, Top , Right, Bottom);
    Rectangle(V_Context, X-Ax, Top , X+Ax , Y-Ay  );
    Rectangle(V_Context, X-Ax, Y+Ay, X+Ax , Bottom);
    Board^.Draw(V_Context,X,Y,Ax,Ay);
  end;
end;

procedure   T_Game.GameOverMessage(V_Winner: boolean);
var Buf: T_GameOverBuf;
begin
  GameOver:= true; Instruct;
  if V_Winner then begin
    if RedPlays then StrCopy(Buf.Winner,'red wins')
                else StrCopy(Buf.Winner,'blue wins');
  end
  else StrCopy(Buf.Winner,'undecided');
  Application^.ExecDialog(new(P_GameOverDlg,Init(@Self,MakeIntResource(DLG_GAMEOVER),@Buf)));
  NewGame;
end;

procedure   T_Game.Instruct;
var Title: array[0..127] of char;
begin
  if GameOver then StrCat(StrECopy(Title,BaseTitle),C_GameOverStr) else
  if RedPlays then StrCat(StrECopy(Title,BaseTitle),C_RedPlaysStr) else
                   StrCat(StrECopy(Title,BaseTitle),C_BluePlaysStr);
  SetCaption(Title);
end;

procedure   T_Game.FirstMove;
begin
  if (Player = Play_Two) or RedPlays then Instruct
  else begin
    DoMove(1 + Random(ColumnCount));
    SwapPlayer;
  end;
end;

procedure T_Game.SwapPlayer;
begin
  RedPlays:= not RedPlays;
  Instruct;
end;

procedure T_Game.DoMove(V_Col: integer);
begin Board^.DoMove(V_Col,RedPlays); end;

procedure   T_Game.WMLButtonDown (var V_Msg: TMessage);
var Col: integer;
begin if Board^.FreeField(V_Msg.LParamLo,V_Msg.LParamHi,Col) then begin
  DoMove(Col);
  case Board^.Scan(Player,RedPlays) of
    Scn_Normal   : if Player = Play_Two then SwapPlayer;
    Scn_Full     : GameOverMessage(false);
    Scn_PrevWins : GameOverMessage(true);
    Scn_NextWins : begin
                     SwapPlayer;
                     if Player <> Play_Two then GameOverMessage(true);
                   end;
  end;
end end;

procedure   T_Game.CMGameOptions  (var V_Msg: TMessage);
var i,n: integer; j,s: T_Player; c,w: T_WhoStarts; Buf: T_OptionsBuf;
begin
  n:= ColumnCount; s:= Player; w:= WhoStarts;
  for i:= C_MinColumns to C_MaxColumns do Buf.ColumnCount[i]:= bf_UnChecked;
  Buf.ColumnCount[n]:= bf_Checked;
  for j:= Play_Red to Play_Two do Buf.Player[j]:= bf_UnChecked;
  Buf.Player[s]:= bf_Checked;
  for c:= Start_Red to Start_Random do Buf.WhoStarts[c]:= bf_UnChecked;
  Buf.WhoStarts[w]:= bf_Checked;
  if Application^.ExecDialog(New(P_OptionsDlg,
    Init(@Self,MakeIntResource(DLG_OPTIONS),@Buf))) = id_OK
  then begin
    for i:= C_MinColumns to C_MaxColumns do if Buf.ColumnCount[i] = bf_Checked then n:= i;
    for j:= Play_Red to Play_Two do if Buf.Player[j] = bf_Checked then s:= j;
    for c:= Start_Red to Start_Random do if Buf.WhoStarts[c] = bf_Checked then w:= c;
    if (n <> ColumnCount) or (s <> Player) or (w <> WhoStarts) then begin
      ColumnCount:= n; Player:= s; WhoStarts:= w;
{$IFDEF DEBUG}
      if not DebugMode then
{$ENDIF}
        NewGame;
    end;
  end;
end;

procedure   T_Game.CMGameNew   (var V_Msg: TMessage);
begin NewGame; end;

procedure   T_Game.CMHelpAbout    (var V_Msg: TMessage);
begin
  MessageBox(HWindow,
    'Four in a Row  (Windows game)'#13#13 +
{$IFDEF DEBUG}
    '<Alt+D> = Debug mode'#13#13 +
{$ENDIF}
    '4IN1ROWW v1.2 (MS-Windows)'#13 +
    '(c) J.R. Ferguson, 1993-2000'#13 +
    'j.r.ferguson@iname.com'#13 +
    'http://hello.to/ferguson',
    'Information',
    mb_IconInformation or mb_OK);
end;

procedure   T_Game.CMHelpRules  (var V_Msg: TMessage);
begin
  MessageBox(HWindow,
    'Red and blue take turns dropping a disc in a column of their choice. ' +
    'Whoever is first to place four discs in a row, wins the game. ' +
    'The winning row may be horizontal, vertical or diagonal.',
    'Rules of the game',
    mb_OK);
end;

{$IFDEF DEBUG}
procedure   T_Game.CMDebug       (var V_Msg: TMessage);
begin
  DebugMode:= MessageBox(HWindow,
              'Continue after changing game options',
              'Debug mode',
              mb_IconQuestion or mb_YesNo)
              = id_Yes;
end;
{$ENDIF}

{ --- T_Application --- }

constructor T_Application.Init(V_Title: PChar);
begin
  inherited Init(V_Title);
  Randomize;
  New(DrawTools,Init);
end;

destructor  T_Application.Done;
begin
  Dispose(DrawTools,Done);
  inherited Done;
end;

procedure   T_Application.InitMainWindow;
begin MainWindow:= New(P_Game, Init(nil,'Four in a Row')); end;

procedure   T_Application.InitInstance;
begin
  inherited InitInstance;
  HAccTable:= LoadAccelerators(HInstance,MakeIntResource(ACC_KEYS));
end;


{ --- Hoofdprogramma --- }

begin
  Application:= New(P_Application,Init('4IN1ROWW'));
  Application^.Run;
  Dispose(Application,Done);
end.
